#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <wininet.h>
#include <winsock2.h>
#include "iTunesVisualAPI.h"

#define VISUALPLUGINNAME	"NowTunes"
#define	VISUALPLUGINCREATOR	0

#define INIFILE			"nowtunes.ini"
#define RESOLUTION		256
#define ALPHA_VALUE		67

#define	VISUALPLUGINMAJORVERSION	0
#define	VISUALPLUGINMINORVERSION	1
#define	VISUALPLUGINRELEASESTAGE	0x80
#define	VISUALPLUGINNONFINALRELEASE	0

#define RVALUE(rgb)	(((rgb) >> 16) & 0xff)
#define GVALUE(rgb)	(((rgb) >>  8) & 0xff)
#define BVALUE(rgb)	(((rgb)      ) & 0xff)
#define AVALUE(s, d, a)	(((d) * (100 - (a))) / 100 + ((s) * (a)) / 100)

typedef struct VisualPluginData {
  void           *appCookie;
  ITAppProcPtr    appProc;

  HWND            hWnd;
  HDC             hDC;
  HBITMAP         hBitmap, hbmpSaved;
  HFONT           hFont, hfontSaved;
  HPEN            hPen, hpenSaved;
  HBRUSH          hbrFg, hbrBg, hbrSaved;
  COLORREF        fg;
  COLORREF        bg;
  int             alpha;
  Rect            destRect;

  ITTrackInfoV1   trackInfo;
  ITStreamInfoV1  streamInfo;

  Boolean         playing;

  RenderVisualData renderData;
  UInt32           timeStampID;
} VisualPluginData;

char url[1024];
char currentSong[1024];

static OSStatus RegisterVisualPlugin(PluginMessageInfo *);
static OSStatus VisualPluginHandler(OSType, VisualPluginMessageInfo *, void *);

static int init(VisualPluginInitMessage *);
static void term(VisualPluginData *);

static int createOffscreen(VisualPluginData *, HWND, Rect *);
static void destroyOffscreen(VisualPluginData *);
static void RenderVisualPort(VisualPluginData *, HWND, const Rect *);

static int checkSong(VisualPluginData *);
static int action(VisualPluginData *);

__declspec(dllexport) OSStatus
iTunesPluginMain(
OSType message,
PluginMessageInfo *vpmi,
void *refCon
)
{
  OSStatus status;
  
  switch (message) {
  case kPluginInitMessage:
    status = RegisterVisualPlugin(vpmi);
    break;
  case kPluginCleanupMessage:
    status = noErr;
    break;
  default:
    status = unimpErr;
    break;
  }

  return status;
}

static OSStatus
RegisterVisualPlugin(
PluginMessageInfo *vpmi
)
{
  OSStatus          status;
  PlayerMessageInfo pmi;
	  
  ZeroMemory(&pmi.u.registerVisualPluginMessage,
	     sizeof(pmi.u.registerVisualPluginMessage));
  
  pmi.u.registerVisualPluginMessage.name[0] = lstrlen(VISUALPLUGINNAME);
  CopyMemory(&pmi.u.registerVisualPluginMessage.name[1], VISUALPLUGINNAME,
	     lstrlen(VISUALPLUGINNAME));

  SetNumVersion(&pmi.u.registerVisualPluginMessage.pluginVersion,
		VISUALPLUGINMAJORVERSION,
		VISUALPLUGINMINORVERSION,
		VISUALPLUGINRELEASESTAGE,
		VISUALPLUGINNONFINALRELEASE);

  pmi.u.registerVisualPluginMessage.handler             = VisualPluginHandler;
  pmi.u.registerVisualPluginMessage.creator             = VISUALPLUGINCREATOR;
  pmi.u.registerVisualPluginMessage.timeBetweenDataInMS = 0xffffffff;
  pmi.u.registerVisualPluginMessage.numWaveformChannels = 0;
  pmi.u.registerVisualPluginMessage.numSpectrumChannels = 2;
  pmi.u.registerVisualPluginMessage.minWidth            = 64;
  pmi.u.registerVisualPluginMessage.minHeight           = 64;
  pmi.u.registerVisualPluginMessage.maxWidth            = 32767;
  pmi.u.registerVisualPluginMessage.maxHeight           = 32767;
  
  status = PlayerRegisterVisualPlugin(vpmi->u.initMessage.appCookie,
				      vpmi->u.initMessage.appProc,
				      &pmi);
	  
  return status;
}

static OSStatus
VisualPluginHandler(
OSType message,
VisualPluginMessageInfo *vpmi,
void *refCon
)
{
  OSStatus          status;
  VisualPluginData *vpd;
  
  vpd = (VisualPluginData *)refCon;
  
  status = noErr;

  switch (message) {
  case kVisualPluginInitMessage:
    if (!init(&vpmi->u.initMessage)) return memFullErr;
    break;
  case kVisualPluginCleanupMessage:
    term(vpd);
    break;
  case kVisualPluginEnableMessage:
  case kVisualPluginDisableMessage:
  case kVisualPluginIdleMessage:
    break;
  case kVisualPluginShowWindowMessage:
    if (createOffscreen(vpd,
			vpmi->u.showWindowMessage.window,
			&vpmi->u.showWindowMessage.drawRect)) {
      RenderVisualPort(vpd, vpd->hWnd, &vpd->destRect);
    }
    break;
  case kVisualPluginSetWindowMessage:
    if (createOffscreen(vpd,
			vpmi->u.setWindowMessage.window,
			&vpmi->u.setWindowMessage.drawRect)) {
      RenderVisualPort(vpd, vpd->hWnd, &vpd->destRect);
    }
    break;
  case kVisualPluginHideWindowMessage:
    destroyOffscreen(vpd);
    break;
  case kVisualPluginRenderMessage:
    CopyMemory(&vpd->renderData, vpmi->u.renderMessage.renderData, sizeof(RenderVisualData));;
    vpd->timeStampID = vpmi->u.renderMessage.timeStampID;
    RenderVisualPort(vpd, vpd->hWnd, &vpd->destRect);
    break;
  case kVisualPluginUpdateMessage:
    RenderVisualPort(vpd, vpd->hWnd, &vpd->destRect);
    break;
  case kVisualPluginPlayMessage:
    if (vpmi->u.playMessage.trackInfo) {
      vpd->trackInfo = *vpmi->u.playMessage.trackInfo;
    } else {
      ZeroMemory(&vpd->trackInfo, sizeof(vpd->trackInfo));
    }

    if (vpmi->u.playMessage.streamInfo) {
      vpd->streamInfo = *vpmi->u.playMessage.streamInfo;
    } else {
      ZeroMemory(&vpd->streamInfo, sizeof(vpd->streamInfo));
    }
    checkSong(vpd);
    vpd->playing = true;
    break;
  case kVisualPluginChangeTrackMessage:
    if (vpmi->u.changeTrackMessage.trackInfo) {
      vpd->trackInfo = *vpmi->u.changeTrackMessage.trackInfo;
    } else {
      ZeroMemory(&vpd->trackInfo, sizeof(vpd->trackInfo));
    }

    if (vpmi->u.changeTrackMessage.streamInfo) {
      vpd->streamInfo = *vpmi->u.changeTrackMessage.streamInfo;
    } else {
      ZeroMemory(&vpd->streamInfo, sizeof(vpd->streamInfo));
    }

    checkSong(vpd);
    break;
  case kVisualPluginStopMessage:
    vpd->playing = false;
    RenderVisualPort(vpd, vpd->hWnd, &vpd->destRect);
    break;
  case kVisualPluginSetPositionMessage:
  case kVisualPluginPauseMessage:
  case kVisualPluginUnpauseMessage:
    break;
  default:
    status = unimpErr;
    break;
  }

  return status;	
}

static int
init(
VisualPluginInitMessage *vpim
)
{
  WSADATA wsaData;
  VisualPluginData *vpd;
  LOGFONT           lf;
  char              buf[256];
  int               r, g, b;

  vpd = (VisualPluginData *)GlobalAlloc(GPTR, sizeof(VisualPluginData));
  if (!vpd) return FALSE;
  ZeroMemory(vpd, sizeof(vpd));
  vpd->appCookie = vpim->appCookie;
  vpd->appProc   = vpim->appProc;
  vpim->refCon= (void *)vpd;

  GetPrivateProfileString("Color", "foreground", "#ffffff", buf, 256, INIFILE);
  sscanf(buf, "#%02x%02x%02x", &r, &g, &b);
  vpd->fg = RGB(r, g, b);
  GetPrivateProfileString("Color", "background", "#000000", buf, 256, INIFILE);
  sscanf(buf, "#%02x%02x%02x", &r, &g, &b);
  vpd->bg = RGB(r, g, b);

  vpd->alpha = GetPrivateProfileInt("Visual", "alpha", ALPHA_VALUE, INIFILE);

  ZeroMemory(&lf, sizeof(LOGFONT));
  GetPrivateProfileString("Font", "face", "FixedSys", lf.lfFaceName,
			  LF_FACESIZE, INIFILE);
  lf.lfHeight = GetPrivateProfileInt("Font", "size", 16, INIFILE) * -1;
  lf.lfWeight = GetPrivateProfileInt("Font", "weight", 0, INIFILE);
  lf.lfCharSet = GetPrivateProfileInt("Font", "charset", 0, INIFILE);

  vpd->hFont = CreateFontIndirect(&lf);
  vpd->hPen  = CreatePen(PS_SOLID, 0,
			 RGB(
			  AVALUE(RVALUE(vpd->fg), RVALUE(vpd->bg), vpd->alpha),
			  AVALUE(GVALUE(vpd->fg), GVALUE(vpd->bg), vpd->alpha),
			  AVALUE(BVALUE(vpd->fg), BVALUE(vpd->bg), vpd->alpha)));
  vpd->hbrFg = CreateSolidBrush(vpd->fg);
  vpd->hbrBg = CreateSolidBrush(vpd->bg);

  GetPrivateProfileString("Action", "URL", "", url, 256, INIFILE);

  WSAStartup(MAKEWORD(2,0), &wsaData);

  return TRUE;
}

static void
term(
VisualPluginData *vpd
)
{
  if (vpd) {
    DeleteObject(vpd->hFont);
    DeleteObject(vpd->hPen);
    DeleteObject(vpd->hbrFg);
    DeleteObject(vpd->hbrBg);
    GlobalFree(vpd);
  }

  WSACleanup();
}

static int
createOffscreen(
VisualPluginData *vpd,
HWND hWnd,
Rect *rc
)
{
  HDC hDC;
  int width, height;

  if (!hWnd) return FALSE;

  hDC = GetDC(hWnd);		
  if (!hDC) return FALSE;

  destroyOffscreen(vpd);

  width  = rc->right - rc->left;
  height = rc->bottom - rc->top;

  vpd->hDC = CreateCompatibleDC(hDC);
  vpd->hBitmap = CreateCompatibleBitmap(hDC, width, height);
  vpd->hbmpSaved = SelectObject(vpd->hDC, vpd->hBitmap);

  SetTextColor(vpd->hDC, vpd->fg);
  SetBkMode(vpd->hDC, TRANSPARENT);
  vpd->hfontSaved = SelectObject(vpd->hDC, vpd->hFont);
  vpd->hpenSaved  = SelectObject(vpd->hDC, vpd->hPen);
  vpd->hbrSaved   = SelectObject(vpd->hDC, vpd->hbrFg);

  ReleaseDC(hWnd, hDC);

  vpd->hWnd = hWnd;
  vpd->destRect = *rc;

  return TRUE;
}

static void
destroyOffscreen(
VisualPluginData *vpd
)
{
  if (vpd->hDC) {
    if (vpd->hbmpSaved) {
      SelectObject(vpd->hDC, vpd->hbmpSaved);
      vpd->hbmpSaved = NULL;
    }
    if (vpd->hBitmap) {
      DeleteObject(vpd->hBitmap);
      vpd->hBitmap = NULL;
    }
    if (vpd->hpenSaved) {
      SelectObject(vpd->hDC, vpd->hpenSaved);
      vpd->hpenSaved = NULL;
    }
    if (vpd->hbrSaved) {
      SelectObject(vpd->hDC, vpd->hbrSaved);
      vpd->hbrSaved = NULL;
    }
    if (vpd->hfontSaved) {
      SelectObject(vpd->hDC, vpd->hfontSaved);
    }
    DeleteDC(vpd->hDC);
    vpd->hDC = NULL;
  }
}

static void
RenderVisualPort(
VisualPluginData *vpd,
HWND hWnd,
const Rect *destRect
)
{
  RECT    srcRect, drawRect;
  HDC     hDC;
  char    buf[1024];
  int     i, x, y, h, r, b;
  UInt8   *spectrum;
	  
  if (!hWnd) return;

  hDC = GetDC(hWnd);		
  if (!hDC) return;

  srcRect.left = srcRect.top = 0;
  srcRect.right = destRect->right - destRect->left;
  srcRect.bottom = destRect->bottom - destRect->top;

  FillRect(vpd->hDC, &srcRect, vpd->hbrBg);

  h = srcRect.bottom >> 1;
  r = srcRect.right;
  b = srcRect.bottom;

  MoveToEx(vpd->hDC, 0, h, NULL);
  spectrum = vpd->renderData.spectrumData[0];
  for (i = 0; i < RESOLUTION; i++) {
    x = (r * i - (RESOLUTION - 1)) / RESOLUTION;
    y = (b * spectrum[i]) >> 9;
    LineTo(vpd->hDC, x, h - y);
  }
  spectrum = vpd->renderData.spectrumData[1];
  for (i = RESOLUTION - 1; i >= 0; i--) {
    x = (r * i - (RESOLUTION - 1)) / RESOLUTION;
    y = (b * spectrum[i]) >> 9;
    LineTo(vpd->hDC, x, h + y);
  }

  if (*vpd->streamInfo.streamURL) {
    wsprintf(buf,
	     "%s",
	     vpd->streamInfo.streamTitle + 1
	     );
  } else if (*vpd->trackInfo.name) {
    wsprintf(buf,
	     "%s/ %s\n"
	     "(%s)",
	     vpd->trackInfo.name + 1,
	     vpd->trackInfo.artist + 1,
	     vpd->trackInfo.album + 1
	     );
  } else {
    *buf = '\0';
  }
  drawRect.left = drawRect.top = drawRect.bottom = 0;
  drawRect.right = srcRect.right;
  DrawText(vpd->hDC, buf, -1, &drawRect, DT_CALCRECT | DT_NOPREFIX | DT_WORDBREAK);
  OffsetRect(&drawRect,
	    (srcRect.right - drawRect.right) >> 1,
	    (srcRect.bottom - drawRect.bottom) >> 1);
  DrawText(vpd->hDC, buf, -1, &drawRect, DT_CENTER | DT_NOPREFIX | DT_WORDBREAK);

  BitBlt(hDC, destRect->left, destRect->top,
	 destRect->right - destRect->left, destRect->bottom - destRect->top,
	 vpd->hDC, srcRect.left, srcRect.top, SRCCOPY);

  ReleaseDC(hWnd, hDC);
}

static
int
checkSong(
VisualPluginData *vpd
)
{
  char buf[1024];

  if (*vpd->streamInfo.streamTitle) {
    wsprintf(buf, "%s (%s)",
	     vpd->streamInfo.streamTitle + 1,
	     vpd->trackInfo.name + 1);
  } else if (*vpd->trackInfo.name) {
    wsprintf(buf, "%s (%s)",
	     vpd->trackInfo.name + 1,
	     vpd->trackInfo.artist + 1);
  } else {
    *buf = '\0';
  }
  if (lstrcmp(buf, currentSong)) {
    lstrcpy(currentSong, buf);
    action(vpd);
  }

  return 1;
}

int
action(
VisualPluginData *vpd
)
{
  char      buf[1024];
  SOCKET    sock;
  struct    sockaddr_in addr;
  BOOL      yes = 1;
  HINTERNET hInet = NULL;
  HINTERNET hHttp = NULL;

  if (*vpd->streamInfo.streamTitle) {
    wsprintf(buf, "song=%s\n", currentSong);
  } else if (*vpd->trackInfo.name) {
    wsprintf(buf, "songtitle=%s\nalbum=%s\nartist=%s\nsong=%s\n",
		  vpd->trackInfo.name + 1,
		  vpd->trackInfo.album + 1,
		  vpd->trackInfo.artist + 1,
		  currentSong);
  }

  sock = socket(AF_INET, SOCK_DGRAM, 0);

  addr.sin_family = AF_INET;
  addr.sin_port = htons(12345);
  addr.sin_addr.S_un.S_addr = inet_addr("255.255.255.255");

  setsockopt(sock, SOL_SOCKET, SO_BROADCAST, (char *)&yes, sizeof(yes));

  sendto(sock, buf, lstrlen(buf), 0, (struct sockaddr *)&addr, sizeof(addr));

  closesocket(sock);

  if (!*url) return 0;

  if (*vpd->streamInfo.streamTitle) {
    wsprintf(buf, "%s?song=%s",
		  url,
		  currentSong);
  } else if (*vpd->trackInfo.name) {
    wsprintf(buf, "%s?songtitle=%s&album=%s&artist=%s&song=%s",
		  url,
		  vpd->trackInfo.name + 1,
		  vpd->trackInfo.album + 1,
		  vpd->trackInfo.artist + 1,
		  currentSong);
  } else {
    return 0;
  }

  hInet = InternetOpen("NowTunes", INTERNET_OPEN_TYPE_DIRECT, NULL, NULL, 0);
  if (!hInet) goto END_OF_FUNC;

  hHttp = InternetOpenUrl(hInet, buf, NULL, -1, 0, 0);
  if (!hHttp) goto END_OF_FUNC;

END_OF_FUNC:
  if (hHttp) InternetCloseHandle(hHttp);
  if (hInet) InternetCloseHandle(hInet);

  return 1;
}

/*
 * vi:ts=8 sw=2
 */
